#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Yikun Zhang
Last Editing: June 19, 2022

Description: This script contains code for the parallel implementations with 
Ray (https://www.ray.io/) of the standard (or Euclidean) kernel density estimator 
with the Gaussian kernel, its gradient estimator, mean shift (MS), and subspace 
constrained mean shift (SCMS) algorithms.
"""

import numpy as np
from numpy import linalg as LA
import ray

#==========================================================================================#

@ray.remote
def KDE_Ray(x, data, h=None, wt=None):
    '''
    The d-dim Euclidean KDE with Gaussian kernel.
    
    Parameters:
        x: (m,d)-array
            The coordinates of m query points in the d-dim Euclidean space.
    
        data: (n,d)-array
            The coordinates of n random sample points in the d-dimensional 
            Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
    
    Return:
        f_hat: (m,)-array
            The corresponding kernel density estimates at m query points.
    '''
    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Dimension of the data
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    f_hat = np.zeros((x.shape[0], ))
    if wt is None:
        wt = np.ones((n,))/n
    for i in range(x.shape[0]):
        f_hat[i] = np.sum(wt*np.exp(np.sum(-((x[i,:] - data)/h)**2, axis=1)/2))/ \
                   ((2*np.pi)**(d/2)*np.prod(h))
    return f_hat


@ray.remote
def MS_Ray(mesh_0, data, h=None, eps=1e-7, max_iter=1000, wt=None):
    '''
    Mean Shift Algorithm with the Gaussian kernel.
    
    Parameters:
        mesh_0: a (m,d)-array
            The coordinates of m initial points in the d-dim Euclidean space.
    
        data: a (n,d)-array
            The coordinates of n data sample points in the d-dim Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
       
        eps: float
            The precision parameter. (Default: eps=1e-7.)
       
        max_iter: int
            The maximum number of iterations for the SCMS algorithm on each 
            initial point. (Default: max_iter=1000)
        
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
    
    Return:
        MS_new: (m,d)-array
            The collection of converged points yielded by the Euclidean SCMS 
            algorithm.
    '''
    
    n = data.shape[0]   ## Number of data points
    d = data.shape[1]   ## Dimension of the data
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], )) 
    if wt is None:
        wt = np.ones((n,))
    else:
        wt = n*wt
    MS_new = np.copy(mesh_0)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            # print('The MS algorithm converges in ' + str(t-1) + 'steps!')
            break
        MS_old = np.copy(MS_new)
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pt = MS_old[i,:]
                ker_w = wt*np.exp(-np.sum(((x_pt-data)/h)**2, axis=1)/2)
                if np.sum(ker_w) == 0:
                    # Set those points with zero density values to NaN
                    nan_arr = np.zeros_like(x_pt)
                    nan_arr[:] = np.nan
                    conv_sign[i] = 1
                    x_new = nan_arr
                else:
                    # Mean shift update
                    x_new = np.sum(data*ker_w.reshape(n,1), axis=0) / np.sum(ker_w)
                    if LA.norm(x_pt - x_new) < eps:
                        conv_sign[i] = 1
                MS_new[i,:] = x_new
            else:
                MS_new[i,:] = MS_old[i,:]
    
    '''
    if t >= max_iter-1:
        print('The MS algorithm reaches the maximum number of iterations,'\
               +str(max_iter)+', and has not yet converged.')
    '''
    MS_new = MS_new[conv_sign == 1]
    # Drop NaN entries before returning the mode candidates
    return MS_new[~np.isnan(MS_new[:,0])]


@ray.remote
def SCMSLog_Ray(mesh_0, data, d=1, h=None, eps=1e-7, max_iter=1000, wt=None,
                stop_cri='proj_grad'):
    '''
    Subspace Constrained Mean Shift algorithm with the log density and Gaussian 
    kernel.
    
    Parameters:
        mesh_0: a (m,D)-array
            The coordinates of m initial points in the D-dim Euclidean space.
    
        data: a (n,D)-array
            The coordinates of n data sample points in the D-dim Euclidean space.
       
        d: int
            The order of the density ridge. (Default: d=1.)
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
       
        eps: float
            The precision parameter. (Default: eps=1e-7.)
       
        max_iter: int
            The maximum number of iterations for the SCMS algorithm on each 
            initial point. (Default: max_iter=1000)
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
       
        stop_cri: string ('proj_grad'/'pts_diff')
            The indicator of which stopping criteria that will be used to 
            terminate the SCMS algorithm. (When stop_cri='pts_diff', the errors 
            between two consecutive iteration points need to be smaller than 
            'eps' for terminating the algorithm. When stop_cri='proj_grad' or 
            others, the projected/principal gradient of the current point need to be 
            smaller than 'eps' for terminating the algorithm.)
            (Default: stop_cri='proj_grad'.)
    
    Return:
        SCMS_new: (m,D)-array
            The collection of converged points yielded by the Euclidean SCMS 
            algorithm.
    '''
    
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Dimension of data points
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    SCMS_new = np.copy(mesh_0)
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))  
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = wt.reshape(n,1)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            # print('The SCMS algorithm converges in ' + str(t-1) + 'steps!')
            break
        SCMS_old = np.copy(SCMS_new)
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pts = SCMS_old[i,:]
                ## Compute the gradient of the log density
                Grad = np.sum(-wt*(x_pts - data) \
                              *np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), axis=0) /(h**2)
                den_prop = np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2))
                if den_prop == 0:
                    # Set those points with zero density values to NaN
                    nan_arr = np.zeros_like(x_pts)
                    nan_arr[:] = np.nan
                    conv_sign[i] = 1
                    x_new = nan_arr
                else:
                    Log_grad = Grad / den_prop
                    ## Compute the Hessian matrix
                    Log_Hess = np.dot((x_pts-data).T, wt*(x_pts-data) \
                                  * np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1)) \
                               /(h**4 * np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2))) \
                    - np.diag(np.ones(len(x_pts,)) / (h**2)) - np.dot(Log_grad.reshape(D,1), Log_grad.reshape(1,D))
                    Log_Hess = np.dot((x_pts-data).T, wt*(x_pts-data) \
                                  * np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1)) /den_prop \
                    - np.diag(np.ones(len(x_pts,)) * (h**2)) - np.dot((h**2)*Log_grad.reshape(D,1), (h**2)*Log_grad.reshape(1,D))
                    ## Spectral decomposition
                    w, v = LA.eig(Log_Hess)
                    ## Obtain the eigenpairs
                    V_d = v[:, np.argsort(w)[:(len(x_pts)-d)]]
                    ## Mean Shift vector
                    ms_v = np.sum(wt*data*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), axis=0) \
                           / np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2)) - x_pts
                    ## Subspace constrained gradient and mean shift vector
                    SCMS_grad = np.dot(V_d, np.dot(V_d.T, Log_grad))
                    SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                    ## SCMS update
                    x_new = SCMS_v + x_pts
                    ## Stopping criteria
                    if stop_cri == 'pts_diff':
                        if LA.norm(SCMS_v) < eps:
                            conv_sign[i] = 1
                    else: 
                        if LA.norm(SCMS_grad) < eps:
                            conv_sign[i] = 1
                    if sum(x_new.imag) > 0:
                        # Remove those points with nonzero imaginary parts
                        nan_arr = np.zeros(x_pts.shape[0], )
                        nan_arr[:] = np.nan
                        conv_sign[i] = 1
                        x_new = nan_arr
                SCMS_new[i,:] = x_new.real
            else:
                SCMS_new[i,:] = SCMS_old[i,:]
        # print(t)
    
    '''
    if t >= max_iter-1:
        print('The SCMS algorithm reaches the maximum number of iterations,'\
              +str(max_iter)+', and has not yet converged.')
    '''
    SCMS_new = SCMS_new[conv_sign == 1]
    # Drop NaN entries before returning the ridge candidates
    return SCMS_new[~np.isnan(SCMS_new[:,0])]



# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: July 03, 2022

# Description: This script contains code for detecting the knots (or intersection 
# points) and computing the length of a given filament/web-like structure.

import numpy as np
from sklearn.cluster import AgglomerativeClustering
from numpy import linalg as LA

def DetectKnot(x, r_in, r_out, fila_map):
    '''
    Detecting if a point on the filament is a knot/intersection 
    (Metric Graph Method).
    
    Parameters
    ----------
        x: (D,)-array
            The coordinate of the point of interest.
        
        r_in: float
            The radius of the inner ball around the point of interest.
        
        r_out: float
            The radius of the outer ball around the point of interest.
        
        fila_map : (N,D)-array
            The coordinates of all the filament points.

    Returns
    -------
        'Knot'/'Non-Knot': str
            The indicator of whether the point of interest is a knot/intersection
            on the input filament.

    '''
    dist = LA.norm(fila_map - x.values, axis=1)
    can_pts = fila_map[(dist >= r_in) & (dist <= r_out)]
    if can_pts.shape[0] < 3:
        return 'Non-Knot'
    clustering = AgglomerativeClustering(n_clusters=None, affinity='euclidean', 
                                         linkage='average', distance_threshold=(r_in + r_out)/2)
    clustering.fit_predict(can_pts)
    if len(np.unique(clustering.labels_)) >= 3:
        return 'Knot'
    else:
        return 'Non-Knot'
    



# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: June 26, 2022

# Description: This script contains code for the directional kernel density 
# estimator with the von Mises kernel, its gradient estimator, directional mean 
# shift (DMS), and directional subspace constrained mean shift (DirSCMS) algorithms.

import numpy as np
from numpy import linalg as LA
import scipy.special as sp

#=================================================================================#

def DirKDE(x, data, h=None, wt=None):
    '''
    The q-dim directional KDE with the von Mises kernel.
    
    Parameters
    ----------
        x: (m,d)-array
            The Eulidean coordinates of m query points on a unit hypersphere, 
            where d=q+1 is the Euclidean dimension of data
    
        data: (n,d)-array
            The Euclidean coordinates of n directional random sample points in 
            the d-dimensional Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has an 
            equal weight "1/n".)
    
    Return
    ----------
        f_hat: (m,)-array
            The corresponding directinal kernel density estimates at m query 
            points.
            
    Example
    --------
    >>> import numpy as np
    >>> from sconce.utils import CirSphSampling
    >>> from sconce.DirSCMS import DirKDE
    >>> cir_samp = CirSphSampling(2000, lat_c=50, lon_range=[-180,180], 
                                  sigma=0.1, pv_ax=np.array([1,0,0]))
    >>> d_hat_dat = DirKDE(cir_samp, cir_samp, h=None)
    >>> d_hat_dat
    '''
    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Euclidean Dimension of the data

    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (d - R_bar ** 2) / (1 - R_bar ** 2)
        if d == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(d / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (d / 2) * (2 * (d - 1) * sp.iv(d/2, 2*kap_hat) + \
                                  (d+1) * kap_hat * sp.iv(d/2+1, 2*kap_hat)))) ** (1/(d + 3))
        print("The current bandwidth is " + str(h) + ".\n")
    
    if wt is None:
        wt = np.ones((n,))/n
    if d == 3:
        f_hat = np.sum((wt * np.exp((np.dot(x, data.T)-1)/(h**2)))/\
                       (2*np.pi*(1-np.exp(-2/h**2))*h**2), axis=1)
    else:
        f_hat = np.sum(wt * np.exp(np.dot(x, data.T)/(h**2))/((2*np.pi)**(d/2)*\
                           sp.iv(d/2-1, 1/(h**2))*h**(d-2)), axis=1)
    return f_hat


def DirKDEGradLog(x, data, h=None, wt=None):
    '''
    The (Riemannian) gradient of the logarithm of the q-dim directional KDE with
    the von Mises kernel.
    
    Parameters
    ----------
        x: (m,d)-array
            The Eulidean coordinates of m query points on a unit hypersphere, 
            where d=q+1 is the Euclidean dimension of data
    
        data: (n,d)-array
            The Euclidean coordinates of n directional random sample points in 
            the d-dimensional Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
    
    Return
    ----------
        Riem_grad_hat: (m,d)-array
            The corresponding gradients of the log directional KDEs at m query 
            points.
    '''
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Euclidean dimension of data points
    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (D - R_bar ** 2) / (1 - R_bar ** 2)
        if D == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(D / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (D / 2) * (2 * (D - 1) * sp.iv(D/2, 2*kap_hat) + \
                                  (D+1) * kap_hat * sp.iv(D/2+1, 2*kap_hat)))) ** (1/(D + 3))
        print("The current bandwidth is " + str(h) + ".\n")
    
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = wt.reshape(n,1)
        
    Riem_grad_hat = np.zeros((x.shape[0], D))
    for i in range(x.shape[0]):
        x_pts = x[i,:].reshape(D,1)
        # Compute the directional KDE up to a constant
        den_prop = np.sum(wt*np.exp((np.dot(data, x_pts)-1)/(h**2)))
        # Compute the total gradient of the log density
        vtot_Log_grad = np.sum(wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)), axis=0) \
            / ((h**2)*den_prop)
        # Compute the projection matrix onto the tangent space 
        proj_mat = np.diag(np.ones(D,)) - np.dot(x_pts, x_pts.T)
        # Riemannian gradient estimator
        Riem_grad_hat[i,:] = np.dot(proj_mat, vtot_Log_grad)   
    return Riem_grad_hat


def DirMS(y_0, data, h=None, eps=1e-7, max_iter=1000, wt=None, diff_method='all'):
    '''
    Directional mean shift algorithm with the von-Mises Kernel.
    
    Parameters
    ----------
        y_0: (N,d)-array
            The Euclidean coordinates of N directional initial points in 
            d-dimensional Euclidean space.
    
        data: (n,d)-array
            The Euclidean coordinates of n directional random sample points in 
            d-dimensional Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
        
        eps: float
            The precision parameter for stopping the mean shift iteration.
        
        max_iter: int
            The maximum number of iterations for the mean shift iteration.
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has the 
            weight "1/n".)
            
        diff_method: str ('all'/'mean')
            The method of computing the differences between two consecutive sets
            of iteration points when they are compared with the precision 
            parameter to stop the algorithm. (When diff_method='all', all the 
            differences between two consecutive sets of iteration points need 
            to be smaller than 'eps' for terminating the algorithm. When 
            diff_method='mean', only the mean difference is compared with 'eps'
            and stop the algorithm.)
    
    Return
    ----------
        MS_path: (N,d,T)-array
            The whole iterative trajectory of every initial point yielded by 
            the DMS algorithm.
    '''

    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Euclidean dimension of the data

    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (d - R_bar ** 2) / (1 - R_bar ** 2)
        if d == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - \
                  2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(d / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (d / 2) * (2 * (d - 1) * sp.iv(d/2, 2*kap_hat) + \
                                  (d+1) * kap_hat * sp.iv(d/2+1, 2*kap_hat)))) ** (1/(d + 3))
        print("The current bandwidth is " + str(h) + ".\n")

    MS_path = np.zeros((y_0.shape[0], d, max_iter))
    MS_path[:,:,0] = y_0
    if wt is None:
        wt = np.ones((n,))
    for t in range(1, max_iter):
        y_can = np.dot(np.exp((np.dot(MS_path[:,:,t-1], data.T)-1)/(h**2)), data * wt.reshape(n,1))
        y_dist = np.sqrt(np.sum(y_can ** 2, axis=1))
        MS_path[y_dist != 0,:,t] = (y_can / y_dist.reshape(len(y_dist), 1))[y_dist != 0]
        MS_path[y_dist == 0,:,t] = MS_path[y_dist == 0,:,t-1]
        if diff_method == 'mean' and \
        np.mean(1- np.diagonal(np.dot(MS_path[:,:,t], MS_path[:,:,t-1].T))) <=eps:
            break
        else:
            if all(1 - np.diagonal(np.dot(MS_path[:,:,t], MS_path[:,:,t-1].T)) <= eps):
                break       

    if t < max_iter-1:
        print('The directional mean shift algorithm converges in ' + str(t) + ' steps!')
    else:
        print('The directional mean shift algorithm reaches the maximum number '\
              'of iterations,' + str(max_iter) + ' and has not yet converged.')
    return MS_path[:,:,:(t+1)]


def DirSCMSLog(mesh_0, data, d=1, h=None, eps=1e-7, max_iter=1000, wt=None,
               stop_cri='proj_grad'):
    '''
    Directional Subspace Constrained Mean Shift algorithm with log density and 
    von-Mises kernel.
    
    Parameters
    ----------
        mesh_0: a (m,D)-array
            The Euclidean coordinates of m directional initial points in the 
            D-dimensional Euclidean space.
    
        data: a (n,D)-array
            The Euclidean coordinates of n directional data sample points in the 
            D-dimensional Euclidean space.
       
        d: int
            The order of the density ridge.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
       
        eps: float
            The precision parameter.
       
        max_iter: int
            The maximum number of iterations for the directional SCMS algorithm 
            on each initial point.
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
            
        stop_cri: string ('proj_grad'/'pts_diff')
            The indicator of which stopping criteria that will be used to 
            terminate the SCMS algorithm. (When stop_cri='pts_diff', the errors 
            between two consecutive iteration points need to be smaller than 
            'eps' for terminating the algorithm. When stop_cri='proj_grad' or 
            others, the projected/principal (Riemannian) gradient of the current 
            point need to be smaller than 'eps' for terminating the algorithm.)
    
    Return
    ----------
        SCMS_path: (m,D,T)-array
            The entire iterative DirSCMS sequence for each initial point.
    '''
    
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Euclidean dimension of data points
    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (D - R_bar ** 2) / (1 - R_bar ** 2)
        if D == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(D / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (D / 2) * (2 * (D - 1) * sp.iv(D/2, 2*kap_hat) + \
                                  (D+1) * kap_hat * sp.iv(D/2+1, 2*kap_hat)))) ** (1/(D + 3))
        print("The current bandwidth is " + str(h) + ".\n")
    
    SCMS_path = np.zeros((mesh_0.shape[0], D, max_iter))
    SCMS_path[:,:,0] = mesh_0
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = n*wt.reshape(n,1)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            print('The directional SCMS algorithm converges in ' + str(t-1) + 'steps!')
            break
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pts = SCMS_path[i,:,t-1]
                x_pts = x_pts.reshape(len(x_pts), 1)
                ## Compute the directional KDE up to a constant
                den_prop = np.sum(wt*np.exp((np.dot(data, x_pts)-1)/(h**2)))
                ## Compute the total gradient of the log density
                vtot_Log_grad = np.sum(wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)), axis=0) \
                                / ((h**2)*den_prop)
                ## Compute the Hessian of the log density 
                Log_Hess = np.dot(data.T, wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)))/((h**4)*den_prop) \
                           - np.dot(vtot_Log_grad.reshape(D,1), vtot_Log_grad.reshape(1,D)) \
                           - np.diag(np.sum(np.dot(wt*data, x_pts) * np.exp((np.dot(data, x_pts)-1)/(h**2))) \
                                     * np.ones(D,))/((h**2)*den_prop)
                proj_mat = np.diag(np.ones(D,)) - np.dot(x_pts, x_pts.T)
                Log_Hess = np.dot(np.dot(proj_mat, Log_Hess), proj_mat)
                w, v = LA.eig(Log_Hess)
                ## Obtain the eigenpairs inside the tangent space
                tang_eig_v = v[:, (abs(np.dot(x_pts.T, v)) < 1e-8)[0,:]]
                tang_eig_w = w[(abs(np.dot(x_pts.T, v)) < 1e-8)[0,:]]
                V_d = tang_eig_v[:, np.argsort(tang_eig_w)[:(x_pts.shape[0]-1-d)]]
                ## Iterative vector for the directional mean shift algorithm
                ms_v = vtot_Log_grad / LA.norm(vtot_Log_grad)
                ## Subspace constrained gradient and mean shift vector
                SCMS_grad = np.dot(V_d, np.dot(V_d.T, vtot_Log_grad))
                SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                ## SCMS update
                x_new = SCMS_v + x_pts.reshape(x_pts.shape[0], )
                x_new = x_new / LA.norm(x_new)
                if LA.norm(SCMS_grad) < eps:
                    conv_sign[i] = 1
                SCMS_path[i,:,t] = x_new
            else:
                SCMS_path[i,:,t] = SCMS_path[i,:,t-1]
        # print(t)
    
    if t >= max_iter-1:
        print('The directional SCMS algorithm reaches the maximum number of '\
              'iterations,'+str(max_iter)+', and has not yet converged.')
    return SCMS_path[:,:,:t]


def DirSCMS(mesh_0, data, d=1, h=None, eps=1e-7, max_iter=1000, wt=None,
            stop_cri='proj_grad'):
    '''
    Directional Subspace Constrained Mean Shift Algorithm with the von-Mises 
    kernel. 
    
    Parameters
    ----------
        mesh_0: a (m,D)-array
            The Euclidean coordinates of m directional initial points in the 
            D-dimensional Euclidean space.
    
        data: a (n,D)-array
            The Euclidean coordinates of n directional data sample points in the 
            D-dimensional Euclidean space.
       
        d: int
            The order of the density ridge.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
       
        eps: float
            The precision parameter.
       
        max_iter: int
            The maximum number of iterations for the directional SCMS algorithm 
            on each initial point.
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
            
        stop_cri: string ('proj_grad'/'pts_diff')
            The indicator of which stopping criteria that will be used to 
            terminate the SCMS algorithm. (When stop_cri='pts_diff', the errors 
            between two consecutive iteration points need to be smaller than 
            'eps' for terminating the algorithm. When stop_cri='proj_grad' or 
            others, the projected/principal (Riemannian) gradient of the current 
            point need to be smaller than 'eps' for terminating the algorithm.)
    
    Return
    ----------
        SCMS_path: (m,D,T)-array
            The entire iterative DirSCMS sequence for each initial point.
    '''
    
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Euclidean dimension of data points
    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (D - R_bar ** 2) / (1 - R_bar ** 2)
        if D == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(D / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (D / 2) * (2 * (D - 1) * sp.iv(D/2, 2*kap_hat) + \
                                  (D+1) * kap_hat * sp.iv(D/2+1, 2*kap_hat)))) ** (1/(D + 3))
        print("The current bandwidth is " + str(h) + ".\n")
    
    SCMS_path = np.zeros((mesh_0.shape[0], D, max_iter))
    SCMS_path[:,:,0] = mesh_0
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))  
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = n*wt.reshape(n,1)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            print('The directional SCMS algorithm converges in ' + str(t-1) + 'steps!')
            break
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pts = SCMS_path[i,:,t-1].reshape(D, 1)
                ## Compute the Hessian matrix
                Hess = np.dot(data.T, wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)))/(h**2) \
                       - np.diag(np.sum(np.dot(wt*data, x_pts) \
                                 * np.exp((np.dot(data, x_pts)-1)/(h**2))) * np.ones(len(x_pts),))
                x_pts = x_pts.reshape(len(x_pts), 1)
                proj_mat = np.diag(np.ones(x_pts.shape[0],)) - np.dot(x_pts, x_pts.T)
                Hess = np.dot(np.dot(proj_mat, Hess), proj_mat)
                w, v = LA.eig(Hess)
                ## Obtain the eigenpairs within the tangent space
                tang_eig_v = v[:, (abs(np.dot(x_pts.T, v)) < 1e-8)[0,:]]
                tang_eig_w = w[(abs(np.dot(x_pts.T, v)) < 1e-8)[0,:]]
                V_d = tang_eig_v[:, np.argsort(tang_eig_w)[:(x_pts.shape[0]-1-d)]]
                vtot_grad = np.sum(wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)), axis=0)
                ## Iterative vector for the directional mean shift algorithm
                ms_v = vtot_grad / LA.norm(vtot_grad)
                ## Subspace constrained gradient and mean shift vector
                SCMS_grad = np.dot(V_d, np.dot(V_d.T, vtot_grad))
                SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                ## SCMS update
                x_new = SCMS_v + x_pts.reshape(x_pts.shape[0], )
                x_new = x_new / LA.norm(x_new)
                ## Stopping criteria
                if stop_cri == 'pts_diff':
                    if LA.norm(SCMS_v) < eps:
                        conv_sign[i] = 1
                else: 
                    if LA.norm(SCMS_grad) < eps:
                        conv_sign[i] = 1
                SCMS_path[i,:,t] = x_new
            else:
                SCMS_path[i,:,t] = SCMS_path[i,:,t-1]
        # print(t)
    
    if t >= max_iter-1:
        print('The directional SCMS algorithm reaches the maximum number of '\
               'iterations,'+str(max_iter)+', and has not yet converged.')
    return SCMS_path[:,:,:t]


# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: June 26, 2022

# Description: This script contains code for the standard (or Euclidean) 
# kernel density estimator with the Gaussian kernel, its gradient estimator, 
# mean shift (MS), and subspace constrained mean shift (SCMS) algorithms.

import numpy as np
from numpy import linalg as LA

#==============================================================================#

def KDE(x, data, h=None, wt=None):
    '''
    The d-dim Euclidean KDE with the Gaussian kernel.
    
    Parameters
    ----------
        x: (m,d)-array
            The coordinates of m query points in the d-dim Euclidean space.
    
        data: (n,d)-array
            The coordinates of n random sample points in the d-dimensional 
            Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
    
    Return
    ----------
        f_hat: (m,)-array
            The corresponding kernel density estimates at m query points.
    '''
    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Dimension of the data
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    f_hat = np.zeros((x.shape[0], ))
    if wt is None:
        wt = np.ones((n,))/n
    for i in range(x.shape[0]):
        f_hat[i] = np.sum(wt*np.exp(np.sum(-((x[i,:] - data)/h)**2, axis=1)/2))/ \
                   ((2*np.pi)**(d/2)*np.prod(h))
    return f_hat


def KDEGradLog(x, data, h=None, wt=None):
    '''
    The gradient of the logarithm of the d-dim Euclidean KDE with Gaussian kernel.
    
    Parameters
    ----------
        x: (m,d)-array
            The coordinates of m query points in the d-dim Euclidean space.
    
        data: (n,d)-array
            The coordinates of n random sample points in the d-dimensional 
            Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
    
    Return
    ----------
        grad_log_hat: (m,d)-array
            The corresponding gradients of the log-KDEs at m query points.
    '''
    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Dimension of the data
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    grad_log_hat = np.zeros((x.shape[0], d))
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = wt.reshape(n,1)
    for i in range(x.shape[0]):
        x_pts = x[i,:]
        # Compute the gradient of the log density
        Grad = np.sum(-wt*(x_pts - data) * \
                      np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), 
                      axis=0) /(h**2)
        grad_log_hat[i,:] = Grad / np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, 
                                                                       axis=1)/2))
    return grad_log_hat



def MS(mesh_0, data, h=None, eps=1e-7, max_iter=1000, wt=None):
    '''
    Mean Shift Algorithm with the Gaussian kernel.
    
    Parameters
    ----------
        mesh_0: a (m,d)-array
            The coordinates of m initial points in the d-dim Euclidean space.
    
        data: a (n,d)-array
            The coordinates of n data sample points in the d-dim Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
       
        eps: float
            The precision parameter. 
       
        max_iter: int
            The maximum number of iterations for the SCMS algorithm on each 
            initial point. 
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
    
    Return
    ----------
        MS_path: (m,d,T)-array
            The entire iterative MS sequence for each initial point.
    '''
    
    n = data.shape[0]   ## Number of data points
    d = data.shape[1]   ## Dimension of the data
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    MS_path = np.zeros((mesh_0.shape[0], d, max_iter))
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], )) 
    if wt is None:
        wt = np.ones((n,))
    else:
        wt = n*wt
    MS_path[:,:,0] = mesh_0
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            print('The MS algorithm converges in ' + str(t-1) + 'steps!')
            break
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pt = MS_path[i,:,t-1]
                ker_w = wt*np.exp(-np.sum(((x_pt-data)/h)**2, axis=1)/2)
                # Mean shift update
                x_new = np.sum(data*ker_w.reshape(n,1), axis=0) / np.sum(ker_w)
                if LA.norm(x_pt - x_new) < eps:
                    conv_sign[i] = 1
                MS_path[i,:,t] = x_new
            else:
                MS_path[i,:,t] = MS_path[i,:,t-1]
    
    if t >= max_iter-1:
        print('The MS algorithm reaches the maximum number of iterations,'\
               +str(max_iter)+', and has not yet converged.')
    return MS_path[:,:,:t]


def SCMSLog(mesh_0, data, d=1, h=None, eps=1e-7, max_iter=1000, wt=None,
                 stop_cri='proj_grad'):
    '''
    Subspace Constrained Mean Shift algorithm with log density and Gaussian kernel.
    
    Parameters
    ----------
        mesh_0: a (m,D)-array
            The coordinates of m initial points in the D-dim Euclidean space.
    
        data: a (n,D)-array
            The coordinates of n data sample points in the D-dim Euclidean space.
       
        d: int
            The order of the density ridge.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
       
        eps: float
            The precision parameter.
       
        max_iter: int
            The maximum number of iterations for the SCMS algorithm on each 
            initial point.
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
       
        stop_cri: string ('proj_grad'/'pts_diff')
            The indicator of which stopping criteria that will be used to 
            terminate the SCMS algorithm. (When stop_cri='pts_diff', the errors 
            between two consecutive iteration points need to be smaller than 
            'eps' for terminating the algorithm. When stop_cri='proj_grad' or 
            others, the projected/principal gradient of the current point need to be 
            smaller than 'eps' for terminating the algorithm.)
    
    Return
    ----------
        SCMS_path: (m,D,T)-array
            The entire iterative SCMS sequence for each initial point.
    '''
    
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Dimension of data points
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    SCMS_path = np.zeros((mesh_0.shape[0], D, max_iter))
    SCMS_path[:,:,0] = mesh_0
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))  
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = wt.reshape(n,1)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            print('The SCMS algorithm converges in ' + str(t-1) + 'steps!')
            break
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pts = SCMS_path[i,:,t-1]
                ## Compute the gradient of the log density
                Grad = np.sum(-wt*(x_pts - data) \
                              *np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), axis=0) /(h**2)
                Log_grad = Grad / np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2))
                ## Compute the Hessian matrix
                Log_Hess = np.dot((x_pts-data).T, wt*(x_pts-data) \
                                  * np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1)) \
                           /(h**4 * np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2))) \
                - np.diag(np.ones(len(x_pts,)) / (h**2)) - np.dot(Log_grad.reshape(D,1), Log_grad.reshape(1,D))
                ## Spectral decomposition
                w, v = LA.eig(Log_Hess)
                ## Obtain the eigenpairs
                V_d = v[:, np.argsort(w)[:(len(x_pts)-d)]]
                ## Mean Shift vector
                ms_v = np.sum(wt*data*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), axis=0) \
                       / np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2)) - x_pts
                ## Subspace constrained gradient and mean shift vector
                SCMS_grad = np.dot(V_d, np.dot(V_d.T, Log_grad))
                SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                ## SCMS update
                x_new = SCMS_v + x_pts
                ## Stopping criteria
                if stop_cri == 'pts_diff':
                    if LA.norm(SCMS_v) < eps:
                        conv_sign[i] = 1
                else: 
                    if LA.norm(SCMS_grad) < eps:
                        conv_sign[i] = 1
                SCMS_path[i,:,t] = x_new
            else:
                SCMS_path[i,:,t] = SCMS_path[i,:,t-1]
        # print(t)
    
    if t >= max_iter-1:
        print('The SCMS algorithm reaches the maximum number of iterations,'\
              +str(max_iter)+', and has not yet converged.')
    return SCMS_path[:,:,:t]


def SCMS(mesh_0, data, d=1, h=None, eps=1e-7, max_iter=1000, wt=None, stop_cri='proj_grad'):
    '''
    Subspace Constrained Mean Shift Algorithm with Gaussian kernel.
    
    Parameters
    ----------
        mesh_0: a (m,D)-array
            The coordinates of m initial points in the D-dim Euclidean space.
    
        data: a (n,D)-array
            The coordinates of n data sample points in the D-dim Euclidean space.
       
        d: int
            The order of the density ridge.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
       
        eps: float
            The precision parameter.
       
        max_iter: int
            The maximum number of iterations for the SCMS algorithm on each 
            initial point.
            
        wt: (n,)-array
            The weights of kernel density contributions for n random sample 
            points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
       
        stop_cri: string ('proj_grad'/'pts_diff')
            The indicator of which stopping criteria that will be used to 
            terminate the SCMS algorithm. (When stop_cri='pts_diff', the errors 
            between two consecutive iteration points need to be smaller than 
            'eps' for terminating the algorithm. When stop_cri='proj_grad' or 
            others, the projected/principal gradient of the current point need to be 
            smaller than 'eps' for terminating the algorithm.)
            
    Return
    ----------
        SCMS_path: (m,D,T)-array
            The entire iterative SCMS sequence for each initial point.
    '''
    
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Dimension of data points
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        # (Only works for Gaussian kernel)
        h = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data, axis=0))
        print("The current bandwidth is "+ str(h) + ".\n")
    
    SCMS_path = np.zeros((mesh_0.shape[0], D, max_iter))
    SCMS_path[:,:,0] = mesh_0
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))  
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = wt.reshape(n,1)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            print('The SCMS algorithm converges in ' + str(t-1) + 'steps!')
            break
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pts = SCMS_path[i,:,t-1]
                ## Compute the Hessian matrix
                Hess = np.dot((x_pts-data).T, wt*(x_pts-data) \
                              *np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1))/(h**2) \
                       - np.diag(np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2)) \
                                 * np.ones(len(x_pts,)))
                ## Spectral decomposition
                w, v = LA.eig(Hess)
                ## Obtain the eigenpairs
                V_d = v[:, np.argsort(w)[:(len(x_pts)-d)]]
                Grad = np.sum(-wt*(x_pts - data) \
                              *np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), axis=0)
                ## Mean Shift vector
                ms_v = np.sum(wt*data*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2).reshape(n,-1), axis=0) \
                       / np.sum(wt.reshape(n,)*np.exp(np.sum(-((x_pts - data)/h)**2, axis=1)/2)) - x_pts
                ## Subspace constrained gradient and mean shift vector
                SCMS_grad = np.dot(V_d, np.dot(V_d.T, Grad))
                SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                ## SCMS update
                x_new = SCMS_v + x_pts
                ## Stopping criteria
                if stop_cri == 'pts_diff':
                    if LA.norm(SCMS_v) < eps:
                        conv_sign[i] = 1
                else: 
                    if LA.norm(SCMS_grad) < eps:
                        conv_sign[i] = 1
                SCMS_path[i,:,t] = x_new
            else:
                SCMS_path[i,:,t] = SCMS_path[i,:,t-1]
        # print(t)
    
    if t >= max_iter-1:
        print('The SCMS algorithm reaches the maximum number of iterations,'\
               +str(max_iter)+', and has not yet converged.')
    return SCMS_path[:,:,:t]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Yikun Zhang
Last Editing: June 19, 2022

Description: This script contains code for the parallel implementations with 
Ray (https://www.ray.io/) of the directional-linear kernel density 
estimator with the von Mises and Gaussian kernels, directional-linear mean 
shift (DirLinMS), and directional-linear subspace constrained mean shift 
(DirLinSCMS) algorithms.
"""

import numpy as np
from numpy import linalg as LA
import scipy.special as sp
import ray

#==========================================================================================#

@ray.remote
def DirLinKDE_Ray(x, data, h=None, b=None, q=2, D=1):
    '''
    Directional-linear KDE with the von Mises and Gaussian kernels
    
    Parameters:
        x: (m, q+1+D)-array
            Eulidean coordinates of m directional-linear query points, where 
            (q+1) is the Euclidean dimension of the directional component (first 
            (q+1) columns) and D is the dimension of the linear component (last 
            D columns).
    
        data: (n, q+1+D)-array
            Euclidean coordinates of n directional-linear random sample points, 
            where (q+1) is the Euclidean dimension of the directional component 
            (first (q+1) columns) and D is the dimension of the linear component
            (last D columns).
       
        h: float
            Bandwidth parameter for the directional component. (Default: h=None. 
            Then a rule of thumb for directional KDEs with the von Mises kernel 
            in Garcia-Portugues (2013) is applied.)
            
        b: float
            Bandwidth parameter for the linear component. (Default: h=None. 
            Then the Silverman's rule of thumb is applied. See Chen et al.(2016) 
            for details.)
            
        q: int
            Intrinsic data dimension of directional components.
            
        D: int
            Data dimension of linear components.
    
    Return:
        f_hat: (m,)-array
            The corresponding directinal-linear density estimates at m query points.
    '''
    n = data.shape[0]  ## Number of data points
    D_t = data.shape[1]  ## Total dimension of the data
    
    assert q+1+D == D_t, "The dimension of the input data, "+str(D_t)+", should "\
    "equal to the sum of directional dimension ("+str(q+1)+") in its ambient "\
    "space and linear dimension ("+str(D)+")."
    
    data_Dir = data[:,:(q+1)]
    x_Dir = x[:,:(q+1)]
    data_Lin = data[:,(q+1):(q+1+D)]
    x_Lin = x[:,(q+1):(q+1+D)]

    # Rule of thumb for directional component
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data_Dir, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (q + 1 - R_bar ** 2) / (1 - R_bar ** 2)
        if q == 2:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv((q-1) / 2 , kap_hat)**2) / \
                 (n * kap_hat ** ((q+1) / 2) * (2 * q * sp.iv((q+1)/2, 2*kap_hat) + \
                                  (q+2) * kap_hat * sp.iv((q+3)/2, 2*kap_hat)))) ** (1/(q + 4))
        print("The current bandwidth for directional component is " + str(h) + ".\n")
    
    # Rule of thumb for linear component
    if b is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter
        b = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data_Lin, axis=0))
        print("The current bandwidth for linear component is "+ str(b) + ".\n")
    
    # Compute the kernel weights contributed by directional components
    if q == 2:
        f_hat_Dir = np.exp((np.dot(x_Dir, data_Dir.T)-1)/(h**2))/(2*np.pi\
                        *(1-np.exp(-2/h**2))*h**2)
    else:
        f_hat_Dir = np.exp(np.dot(x_Dir, data_Dir.T)/(h**2))/((2*np.pi)**((q+1)/2)*\
                           sp.iv((q-1)/2, 1/(h**2))*h**(q-1))
    
    # Compute the kernel weights contributed by linear components
    f_hat_Lin = np.zeros((x.shape[0], n))
    for i in range(x.shape[0]):
        f_hat_Lin[i,:] = np.exp(np.sum(-((x_Lin[i,:] - data_Lin)/b)**2, axis=1)/2)/ \
                        ((2*np.pi)**(D/2)*np.prod(b))
    
    f_hat = np.mean(f_hat_Dir * f_hat_Lin, axis=1)
    return f_hat


@ray.remote
def DirLinMS_Ray(mesh_0, data, h=None, b=None, q=2, D=1, eps=1e-7, max_iter=1000):
    '''
    Directional-linear Mean Shift Algorithm with the von Mises and Gaussian kernels
    (Simultaneous version).
    
    Parameters:
        mesh_0: (m, q+1+D)-array
            Eulidean coordinates of m directional-linear query points, where 
            (q+1) is the Euclidean dimension of the directional component (first 
            (q+1) columns) and D is the dimension of the linear component (last 
            D columns).
    
        data: (n, q+1+D)-array
            Euclidean coordinates of n directional-linear random sample points, 
            where (q+1) is the Euclidean dimension of the directional component 
            (first (q+1) columns) and D is the dimension of the linear component
            (last D columns).
       
        h: float
            Bandwidth parameter for the directional component. (Default: h=None. 
            Then a rule of thumb for directional KDEs with the von Mises kernel 
            in Garcia-Portugues (2013) is applied.)
            
        b: float
            Bandwidth parameter for the linear component. (Default: h=None. 
            Then the Silverman's rule of thumb is applied. See Chen et al.(2016) 
            for details.)
            
        q: int
            Intrinsic data dimension of directional components. (Default: q=2.)
            
        D: int
            Data dimension of linear components. (Default: D=1.)
       
        eps: float
            The precision parameter. (Default: eps=1e-7.)
       
        max_iter: int
            The maximum number of iterations for the MS algorithm on each 
            initial point. (Default: max_iter=1000.)
    
    Return:
        MS_new: (m,q+1+D)-array
            The collection of converged points yielded by the DirLinMS algorithm.
    '''
    
    n = data.shape[0]  ## Number of data points
    D_t = data.shape[1]  ## Total dimension of the data
    
    assert q+1+D == D_t, "The dimension of the input data, "+str(D_t)+", should "\
    "equal to the sum of directional dimension ("+str(q+1)+") in its ambient "\
    "space and linear dimension ("+str(D)+")."
    
    data_Dir = data[:,:(q+1)]
    data_Lin = data[:,(q+1):(q+1+D)]

    # Rule of thumb for directional component
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data_Dir, axis=0) ** 2))
        # An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (q + 1 - R_bar ** 2) / (1 - R_bar ** 2)
        if q == 2:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv((q-1) / 2 , kap_hat)**2) / \
                 (n * kap_hat ** ((q+1) / 2) * (2 * q * sp.iv((q+1)/2, 2*kap_hat) + \
                                  (q+2) * kap_hat * sp.iv((q+3)/2, 2*kap_hat)))) ** (1/(q + 4))
        print("The current bandwidth for directional component is " + str(h) + ".\n")
    
    # Rule of thumb for linear component
    if b is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        b = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data_Lin, axis=0))
        print("The current bandwidth for linear component is "+ str(b) + ".\n")
    
    MS_new = np.copy(mesh_0)
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], )) 
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            # print('The directional-linear MS algorithm converges in ' + str(t-1) + 'steps!')
            break
        MS_old = np.copy(MS_new)
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_Dir = MS_old[i,:(q+1)]
                x_Lin = MS_old[i,(q+1):(q+1+D)]
                # Kernel weights
                ker_w_Dir = np.exp((np.dot(data_Dir, x_Dir) - 1)/ h**2)
                ker_w_Lin = np.exp(-np.sum(((x_Lin - data_Lin)/b)**2, axis=1)/2)
                # Mean shift updates for directional components
                x_Dir_new = np.sum(data_Dir * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1), 
                                   axis=0)
                x_Dir_new = x_Dir_new / LA.norm(x_Dir_new)
                MS_new[i,:(q+1)] = x_Dir_new
                # Mean shift updates for linear components
                x_Lin_new = np.sum(data_Lin * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1),
                                   axis=0) / np.sum(ker_w_Dir * ker_w_Lin)
                MS_new[i,(q+1):(q+1+D)] = x_Lin_new
                if LA.norm(MS_new - MS_old) <= eps:
                    conv_sign[i] = 1
            else:
                MS_new[i,:] = MS_old[i,:]
                
    if t >= max_iter-1:
        print('The MS algorithm reaches the maximum number of iterations,'\
               +str(max_iter)+', and has not yet converged.')
    MS_new = MS_new[conv_sign == 1]
    # Drop NaN entries before returning the mode candidates
    return MS_new[~np.isnan(MS_new[:,0])]

                
@ray.remote             
def DirLinMS_CA_Ray(mesh_0, data, h=None, b=None, q=2, D=1, eps=1e-7, max_iter=1000):
    '''
    Directional-Linear Mean Shift Algorithm with the von Mises and Gaussian kernels
    (Componentwise version).
    
    Parameters:
        mesh_0: (m, q+1+D)-array
            Eulidean coordinates of m directional-linear query points, where 
            (q+1) is the Euclidean dimension of the directional component (first 
            (q+1) columns) and D is the dimension of the linear component (last 
            D columns).
    
        data: (n, q+1+D)-array
            Euclidean coordinates of n directional-linear random sample points, 
            where (q+1) is the Euclidean dimension of the directional component 
            (first (q+1) columns) and D is the dimension of the linear component
            (last D columns).
       
        h: float
            Bandwidth parameter for the directional component. (Default: h=None. 
            Then a rule of thumb for directional KDEs with the von Mises kernel 
            in Garcia-Portugues (2013) is applied.)
            
        b: float
            Bandwidth parameter for the linear component. (Default: h=None. 
            Then the Silverman's rule of thumb is applied. See Chen et al.(2016) 
            for details.)
            
        q: int
            Intrinsic data dimension of directional components. (Default: q=2.)
            
        D: int
            Data dimension of linear components. (Default: D=1.)
       
        eps: float
            The precision parameter. (Default: eps=1e-7.)
       
        max_iter: int
            The maximum number of iterations for the MS algorithm on each 
            initial point. (Default: max_iter=1000.)
    
    Return:
        MS_new: (m,q+1+D)-array
            The collection of converged points yielded by the DirLinMS algorithm.
    '''
    
    n = data.shape[0]  ## Number of data points
    D_t = data.shape[1]  ## Total dimension of the data
    
    assert q+1+D == D_t, "The dimension of the input data, "+str(D_t)+", should "\
    "equal to the sum of directional dimension ("+str(q+1)+") in its ambient "\
    "space and linear dimension ("+str(D)+")."
    
    data_Dir = data[:,:(q+1)]
    data_Lin = data[:,(q+1):(q+1+D)]

    # Rule of thumb for directional component
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data_Dir, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (q + 1 - R_bar ** 2) / (1 - R_bar ** 2)
        if q == 2:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv((q-1) / 2 , kap_hat)**2) / \
                 (n * kap_hat ** ((q+1) / 2) * (2 * q * sp.iv((q+1)/2, 2*kap_hat) + \
                                  (q+2) * kap_hat * sp.iv((q+3)/2, 2*kap_hat)))) ** (1/(q + 4))
        print("The current bandwidth for directional component is " + str(h) + ".\n")
    
    # Rule of thumb for linear component
    if b is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        b = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data_Lin, axis=0))
        print("The current bandwidth for linear component is "+ str(b) + ".\n")
    
    MS_new = np.copy(mesh_0)
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], )) 
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            # print('The directional-linear MS algorithm converges in '+ str(t-1) + ' steps!')
            break
        MS_old = np.copy(MS_new)
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_Dir = MS_old[i,:(q+1)]
                x_Lin = MS_old[i,(q+1):(q+1+D)]
                # Kernel weights
                ker_w_Dir = np.exp((np.dot(data_Dir, x_Dir) - 1)/ h**2)
                ker_w_Lin = np.exp(-np.sum(((x_Lin - data_Lin)/b)**2, axis=1)/2)
                # Mean shift updates for directional components
                x_Dir_new = np.sum(data_Dir * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1), 
                                   axis=0)
                x_Dir_new = x_Dir_new / LA.norm(x_Dir_new)
                MS_new[i,:(q+1)] = x_Dir_new
                # Recompute the directional kernel weights
                ker_w_Dir = np.exp((np.dot(x_Dir_new, data_Dir.T) - 1)/ h**2)
                # Mean shift updates for linear components
                x_Lin_new = np.sum(data_Lin * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1),
                                   axis=0) / np.sum(ker_w_Dir * ker_w_Lin)
                MS_new[i,(q+1):(q+1+D)] = x_Lin_new
                if LA.norm(MS_new - MS_old) <= eps:
                    conv_sign[i] = 1
            else:
                MS_new[i,:] = MS_old[i,:]
                
    if t >= max_iter-1:
        print('The MS algorithm reaches the maximum number of iterations,'\
               +str(max_iter)+', and has not yet converged.')
    MS_new = MS_new[conv_sign == 1]
    # Drop NaN entries before returning the mode candidates
    return MS_new[~np.isnan(MS_new[:,0])]


@ray.remote
def DirLinSCMSLog_Ray(mesh_0, data, d=1, h=None, b=None, q=2, D=1, eps=1e-7, 
                      max_iter=1000, return_conv=False):
    '''
    Directional-linear Subspace Constrained Mean Shift Algorithm under the 
    log-density with the von Mises and Gaussian kernels (Simultaneous version)
    
    Parameters:
        mesh_0: (m, q+1+D)-array
            Eulidean coordinates of m directional-linear query points, where 
            (q+1) is the Euclidean dimension of the directional component (first 
            (q+1) columns) and D is the dimension of the linear component (last 
            D columns).
    
        data: (n, q+1+D)-array
            Euclidean coordinates of n directional-linear random sample points, 
            where (q+1) is the Euclidean dimension of the directional component 
            (first (q+1) columns) and D is the dimension of the linear component
            (last D columns).
       
        h: float
            Bandwidth parameter for the directional component. (Default: h=None. 
            Then a rule of thumb for directional KDEs with the von Mises kernel 
            in Garcia-Portugues (2013) is applied.)
            
        b: float
            Bandwidth parameter for the linear component. (Default: h=None. 
            Then the Silverman's rule of thumb is applied. See Chen et al.(2016) 
            for details.)
            
        q: int
            Intrinsic data dimension of directional components. (Default: q=2.)
            
        D: int
            Data dimension of linear components. (Default: D=1.)
       
        eps: float
            The precision parameter. (Default: eps=1e-7.)
       
        max_iter: int
            The maximum number of iterations for the MS algorithm on each 
            initial point. (Default: max_iter=1000.)
        
        return_conv: Boolean
            The indicator of whether the convergence signs for the initial points
            will be returned or not.
            
    Return:
        SCMS_new: (m,q+1+D)-array
            The collection of converged points yielded by the DirLinSCMS algorithm.
    '''
    
    n = data.shape[0]  ## Number of data points
    D_t = data.shape[1]  ## Total dimension of the data
    
    assert q+1+D == D_t, "The dimension of the input data, "+str(D_t)+", should "\
    "equal to the sum of directional dimension ("+str(q+1)+") in its ambient "\
    "space and linear dimension ("+str(D)+")."
    
    data_Dir = data[:,:(q+1)]
    data_Lin = data[:,(q+1):(q+1+D)]

    # Rule of thumb for directional component
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data_Dir, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (q + 1 - R_bar ** 2) / (1 - R_bar ** 2)
        if q == 2:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv((q-1) / 2 , kap_hat)**2) / \
                 (n * kap_hat ** ((q+1) / 2) * (2 * q * sp.iv((q+1)/2, 2*kap_hat) + \
                                  (q+2) * kap_hat * sp.iv((q+3)/2, 2*kap_hat)))) ** (1/(q + 4))
        print("The current bandwidth for directional component is " + str(h) + ".\n")
    
    # Rule of thumb for linear component
    if b is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        b = (4/(D+2))**(1/(D+4))*(n**(-1/(D+4)))*np.mean(np.std(data_Lin, axis=0))
        print("The current bandwidth for linear component is "+ str(b) + ".\n")
    
    SCMS_new = np.copy(mesh_0)
    # Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))
    
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            # print('The directional-linear SCMS algorithm converges in ' + str(t-1) + ' steps!')
            break
        SCMS_old = np.copy(SCMS_new)
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_Dir = SCMS_old[i,:(q+1)]
                x_Lin = SCMS_old[i,(q+1):(q+1+D)]
                x_pts = SCMS_old[i,:]
                # Kernel weights
                ker_w_Dir = np.exp((np.dot(data_Dir, x_Dir) - 1)/ h**2)
                ker_w_Lin = np.exp(-np.sum(((x_Lin - data_Lin)/b)**2, axis=1)/2)
                # Compute the directional-linear KDE up to a constant
                den_prop = np.sum(ker_w_Dir * ker_w_Lin)
                if den_prop == 0:
                    # Set those points with zero density values to NaN
                    nan_arr = np.zeros_like(x_pts)
                    nan_arr[:] = np.nan
                    conv_sign[i] = 1
                    x_new = nan_arr
                else:
                    # Compute the total gradient of the log density
                    tot_grad_Log_Dir = np.sum(data_Dir * ker_w_Dir.reshape(n,1) \
                                             * ker_w_Lin.reshape(n,1), axis=0)/((h**2)*den_prop)
                    tot_grad_Log_Lin = np.sum((data_Lin - x_Lin) * ker_w_Dir.reshape(n,1) \
                                               * ker_w_Lin.reshape(n,1), axis=0)/((b**2)*den_prop)
                    tot_grad_Log = np.concatenate([tot_grad_Log_Dir.reshape(q+1, 1), 
                                                   tot_grad_Log_Lin.reshape(D, 1)], axis=0)
                    # Compute the Hessian matrix of the log density
                    ## Hessian in the directional component: (q+1)*(q+1)
                    Log_Hess_Dir = np.dot(data_Dir.T, data_Dir * ker_w_Dir.reshape(n,1) \
                                         * ker_w_Lin.reshape(n,1))/((h**4)*den_prop) \
                        - np.dot(tot_grad_Log_Dir.reshape(q+1, 1), tot_grad_Log_Dir.reshape(1, q+1)) \
                        - np.eye(q+1) * np.sum(np.dot(data_Dir * ker_w_Dir.reshape(n,1) \
                                                      * ker_w_Lin.reshape(n,1), x_Dir))/((h**2)*den_prop)
                    x_Dir = x_Dir.reshape(q+1, 1)
                    proj_mat = np.eye(q+1) - np.dot(x_Dir, x_Dir.T)
                    Log_Hess_Dir = np.dot(np.dot(proj_mat, Log_Hess_Dir), proj_mat)
                    ## Hessian in the linear component: D*D
                    Log_Hess_Lin = np.dot((x_Lin - data_Lin).T, (x_Lin - data_Lin) \
                                * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1))/((b**4)*den_prop) \
                              - np.eye(D) * np.sum(ker_w_Dir.reshape(n,1) \
                                                   * ker_w_Lin.reshape(n,1))/((b**2)*den_prop) \
                              - np.dot(tot_grad_Log_Lin.reshape(D,1), tot_grad_Log_Lin.reshape(1,D))
                    ## Hessian in the off-diagonal part: (q+1)*D
                    x_Dir = x_Dir.reshape(q+1, )
                    Log_Hess_Off = np.dot(data_Dir.T, (data_Lin - x_Lin) * ker_w_Dir.reshape(n,1) \
                                       * ker_w_Lin.reshape(n,1))/((b**2)*(h**2)*den_prop) \
                             - np.dot(tot_grad_Log_Dir.reshape(q+1, 1), tot_grad_Log_Lin.reshape(1,D))
                    Log_Hess_Off = np.dot(proj_mat, Log_Hess_Off)
                    ## Concatenate to obtain the final Hessian
                    Log_Hess = np.zeros((q+1+D, q+1+D))
                    Log_Hess[:(q+1), :(q+1)] = Log_Hess_Dir
                    Log_Hess[:(q+1), (q+1):(q+1+D)] = Log_Hess_Off
                    Log_Hess[(q+1):(q+1+D), :(q+1)] = Log_Hess_Off.T
                    Log_Hess[(q+1):(q+1+D), (q+1):(q+1+D)] = Log_Hess_Lin
                    if (np.sum(np.isinf(Log_Hess)) > 0) or (np.sum(np.isnan(Log_Hess)) > 0):
                        # Set those points with infinite Hessians to NaN
                        nan_arr = np.zeros_like(x_pts)
                        nan_arr[:] = np.nan
                        conv_sign[i] = 1
                        x_new = nan_arr
                    else:
                        # Spectral decomposition
                        w, v = LA.eig(Log_Hess)
                        x_eig = np.concatenate([x_Dir.reshape(q+1, 1), 
                                                np.zeros((D,1))], axis=0)
                        # Obtain the eigenpairs within the tangent space
                        tang_eig_v = v[:, (abs(np.dot(x_eig.T, v)) < 1e-8)[0,:]]
                        tang_eig_w = w[(abs(np.dot(x_eig.T, v)) < 1e-8)[0,:]]
                        V_d = tang_eig_v[:, np.argsort(tang_eig_w)[:(q+D-d)]]
                        # Mean shift vector in directional and linear components
                        ms_Dir = (np.sum(data_Dir * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1), 
                                        axis=0) / np.sum(ker_w_Dir * ker_w_Lin)) * np.min([b/h, 1/(h**2)])
                        ms_Lin = (np.sum(data_Lin * ker_w_Dir.reshape(n,1) * ker_w_Lin.reshape(n,1),
                                        axis=0) / np.sum(ker_w_Dir * ker_w_Lin) - x_Lin) * np.min([h/b, 1/(b**2)])
                        ms_v = np.concatenate([ms_Dir.reshape(q+1, 1),
                                               ms_Lin.reshape(D, 1)], axis=0)
                        # Subspace constrained gradient and mean shift vector
                        SCMS_grad = np.dot(V_d, np.dot(V_d.T, tot_grad_Log))
                        # SCMS_grad_skew = np.dot(V_d, np.dot(V_d.T, tot_grad_Log_skew))
                        SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                        ## SCMS update
                        x_new = SCMS_v + x_pts.reshape(q+1+D, 1)
                        x_new = x_new.reshape(q+1+D, )
                        x_new[:(q+1)] = x_new[:(q+1)] / LA.norm(x_new[:(q+1)])
                        # if (LA.norm(SCMS_grad) < eps) or (LA.norm(SCMS_grad_skew) < eps):
                        if LA.norm(SCMS_grad) < eps:
                            conv_sign[i] = 1
                SCMS_new[i,:] = x_new
            else:
                SCMS_new[i,:] = SCMS_old[i,:]
    
    if t >= max_iter-1:
        print('The directional-linear SCMS algorithm reaches the maximum number of '\
               'iterations,'+str(max_iter)+', and has not yet converged.')
    if return_conv:
        conv_sign = conv_sign[~np.isnan(SCMS_new[:,0])]
        res = np.concatenate([SCMS_new[~np.isnan(SCMS_new[:,0])], conv_sign.reshape(-1,1)], axis=1)
        return res
    # Keep only the converged points
    SCMS_new = SCMS_new[conv_sign == 1]
    return SCMS_new[~np.isnan(SCMS_new[:,0])]



# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: June 26, 2022

# Description: This script contains the utility functions for using our package 
# in practice.

import numpy as np
from numpy import linalg as LA

#================================================================================#

def cart2sph(x, y, z):
    '''
    Converting the Euclidean coordinate of a data point in R^3 to its Spherical 
    coordinates.
    
    Parameters
    ----------
        x, y, z: floats 
            Euclidean coordinate in R^3 of a data point.
    
    Returns
    ----------
        theta: float
            Longitude (ranging from -180 degree to 180 degree).
        phi: float
            Latitude (ranging from -90 degree to 90 degree).
        r: float 
            Radial distance from the origin to the data point.
    '''
    dxy = np.sqrt(x**2 + y**2)
    r = np.sqrt(dxy**2 + z**2)
    theta = np.arctan2(y, x)
    phi = np.arctan2(z, dxy)
    theta, phi = np.rad2deg([theta, phi])
    return theta, phi, r

def sph2cart(theta, phi, r=1):
    '''
    Converting the Spherical coordinate of a data point to its Euclidean 
    coordinate in R^3.
    
    Parameters
    ----------
        theta: float
            Longitude (ranging from -180 degree to 180 degree).
        phi: float
            Latitude (ranging from -90 degree to 90 degree).
        r: float 
            Radial distance from the origin to the data point.
            
    Returns
    ----------
        x, y, z: floats 
            Euclidean coordinate in R^3 of a data point.
    '''
    theta, phi = np.deg2rad([theta, phi])
    z = r * np.sin(phi)
    rcosphi = r * np.cos(phi)
    x = rcosphi * np.cos(theta)
    y = rcosphi * np.sin(theta)
    return x, y, z

def CirSphSampling(N, lat_c=60, lon_range=[-180,180], sigma=0.01, 
                   pv_ax=np.array([0,0,1])):
    '''
    Generating data points from a circle on the unit sphere with additive 
    Gaussian noises to their Cartesian coordinates plus L2 normalizations.
    
    Parameters
    ----------
        N: int
            The number of randomly generated data points.
            
        lat_c: float
            The latitude of the circle with respect to the pivotal axis. 
            (range: 0-90)
            
        lon_range: 2-element list
            The longitude range that the circular structure covers. When 
            "lon_range=[-180,180]", the underlying structure is a full circle.
            
        sigma: float
            The standard deviation of Gaussian noises.
    
        pv_ax: (3,)-array
            The pivotal axis of the circle on the sphere from which the data 
            points are generated (plus noises).
            
    Return
    ----------
        pts_c_noise: (N,3)-array
            The Cartesian coordinates of N simulated data points.
    
    '''
    ## Random longitudes with range (0, 180)
    lon_c = np.random.rand(N,)*(lon_range[1]-lon_range[0]) + lon_range[0]
    lat_c = np.ones((N,))*lat_c
    x_c, y_c, z_c = sph2cart(lon_c, lat_c)

    pts_c = np.concatenate((x_c.reshape(len(x_c), 1), 
                            y_c.reshape(len(y_c), 1),
                            z_c.reshape(len(z_c), 1)), axis=1)
    ## Add Gaussian noises
    pts_c_noise = pts_c + sigma * np.random.randn(pts_c.shape[0], pts_c.shape[1])
    ## Standardize the noisy points
    pts_c_noise = pts_c_noise/np.sqrt(np.sum(pts_c_noise**2, axis=1)).reshape(N,1)
    
    ## Rotate the data samples accordingly
    mu_c = np.array([[0,0,1]])
    R = 2*np.dot(pv_ax.reshape(3,1)+mu_c.T, pv_ax.reshape(1,3)+mu_c)/\
        np.sum((mu_c+pv_ax.reshape(1,3))**2, axis=1) - np.identity(3)
    pts_c_noise = np.dot(R, pts_c_noise.T).T
    return pts_c_noise


def GaussMixture(N, mu=np.array([[1,1]]), cov=np.diag([1,1]).reshape(2,2,1), 
                 prob=[1.0]):
    '''
    Generating data points from a Gaussian mixture model.
    
    Parameters
    ----------
        N: int
            The number of randomly generated data points.
    
        mu: (m,d)-array
            The means of the Gaussian mixture model with m components.
       
        cov: (d,d,m)-array
            The (d,d)-covariance matrices of the Gaussian mixture model with 
            m components.
            
        prob: list of floats
            The mixture probabilities.
    
    Return
    ----------
        data_ps: (N,d)-array
            The Cartesian coordinates of N simulated data points.
    '''
    m = len(prob)   ## The number of mixtures
    d = mu.shape[1]  ## Dimension of the data
    assert (cov.shape[2] == len(prob)), "'cov.shape[2]' and 'len(prob)' "\
    "should be equal."
    inds = np.random.choice(list(range(m)), N, replace=True, 
                            p=np.array(prob)/sum(prob))
    data_ps = np.zeros((N,d))
    for i in range(m):
        data_ps[inds == i,:] = np.random.multivariate_normal(mu[i,:], cov[:,:,i], 
                                                             size=sum(inds == i))
    return data_ps


def vMFDensity(x, mu=np.array([[0,0,1]]), kappa=[1.0], prob=[1.0]):
    '''
    The q-dimensional von-Mises Fisher (vMF) density function or its mixture.
    
    Parameters
    ----------
        x: (n,d)-array
            The Eulidean coordinates of n query points on a unit hypersphere, 
            where d=q+1 is the Euclidean dimension of data.
    
        mu: (m,d)-array
            The Euclidean coordinates of the m mean directions for a mixture of 
            vMF densities.
       
        kappa: list of floats
            The concentration parameters for a mixture of vMF densities.
       
        prob: list of floats
            The mixture probabilities.
            
    Return
    --------
        mix_den: (n,)-array
            The corresponding density value on each query point.
    '''
    assert (mu.shape[1] == x.shape[1] and mu.shape[0] == len(prob)), \
    "The parameter 'x' and mu' should be a (n,d)-array and (m,d)-array, respectively, \
    and 'prob' should be a list of length m."
    assert (len(kappa) == len(prob)), "The parameters 'kappa' and 'prob' should \
    be of the same length."
    d = x.shape[1]   ## Euclidean dimension of the data
    prob = np.array(prob).reshape(len(prob), 1)
    kappa = np.array(kappa)
    dens = kappa**(d/2-1)*np.exp(kappa*np.dot(x, mu.T))/((2*np.pi)**(d/2)*sp.iv(d/2-1, kappa))
    mix_den = np.dot(dens, prob)
    return mix_den


def vMFSamp(n, mu=np.array([0,0,1]), kappa=1):
    '''
    Randomly sampling data points from a q-dimensional von-Mises Fisher (vMF) 
    density with analytic approaches.
    
    Parameters
    ---------
        n: int
            The number of sampling random data points.
        
        mu: (d, )-array
            The Euclidean coordinate of the mean directions of the q-dim vMF
            density, where d=q+1.
            
        kappa: float
            The concentration parameter of the vMF density.
    
    Return
    --------
        data_ps: (n, d)-array
            The Euclidean coordinates of the randomly sampled points from the 
            vMF density.
    '''
    
    d = len(mu)   # Euclidean dimension of the data
    if d == 3:    # when d=3 (S^2 case), no rejection sampling is required
        U1 = np.random.rand(n,1)
        U2 = 2*np.pi*np.random.rand(n,1)
        W = 1 + np.log(U1 + (1-U1)*np.exp(-2*kappa))/kappa
        X = np.cos(U2) * np.sqrt(1-W**2)
        Y = np.sin(U2) * np.sqrt(1-W**2)
        data_ps = np.concatenate((X,Y,W), axis=1)
    else:
        t1 = np.sqrt(4*(kappa**2) + (d-1)**2)
        b = (-2*kappa + t1)/(d-1)
        x0 = (1-b)/(1+b)
        m = (d-1)/2
        c = kappa*x0 + (d-1)*np.log(1-x0**2)
        data_ps = np.zeros((n,d))
        for i in range(n):
            t = -1000
            U = 1
            while t < np.log(U) + c:
                z = np.random.beta(a=m, b=m)
                U = np.random.rand(1,)[0]
                w = (1-(1+b)*z)/(1-(1-b)*z)
                t = kappa*w + (d-1)*np.log(1-x0*w)
            # Generate a vector from the uniform distribution on the unit hypersphere
            V = np.random.multivariate_normal(mean=np.zeros((d-1,)), 
                                              cov=np.identity(d-1), size=1)[0]
            V = V/np.sqrt(np.sum(V**2))
            data_ps[i, 0:(d-1)] = np.sqrt(1-w**2)*V
            data_ps[i, d-1] = w
    # Rotate the data samples accordingly
    mu_c = np.zeros((1,d))
    mu_c[0,d-1] = 1
    R = 2*np.dot(mu.reshape(d,1)+mu_c.T, mu.reshape(1,d)+mu_c) / \
        np.sum((mu_c+mu.reshape(1,d))**2, axis=1) - np.identity(d)
    data_ps = np.dot(R, data_ps.T).T
    return data_ps


def vMFRejectSamp(n, mu=np.array([0,0,1]), kappa=1):
    '''
    Randomly sampling data points from a q-dimensional von-Mises Fisher (vMF) 
    density via rejection sampling.
    
    Parameters
    ---------
        n: int
            The number of sampling random data points.
        
        mu: (d, )-array
            The Euclidean coordinate of the mean directions of the q-dim vMF
            density, where d=q+1.
            
        kappa: float
            The concentration parameter of the vMF density.
    
    Return
    --------
        data_ps: (n, d)-array
            The Euclidean coordinates of the randomly sampled points from the 
            vMF density.
    '''
    d = len(mu)   ## Euclidean dimension of the data
    data_ps = np.zeros((n,d))
    ## Sample points from standard normal and then standardize them
    sam_can = np.random.multivariate_normal(mean=np.zeros((d,)), cov=np.identity(d), size=n)
    dist_sam = np.sqrt(np.sum(sam_can**2, axis=1)).reshape(n,1)
    sam_can = sam_can/dist_sam

    unif_sam = np.random.uniform(0, 1, n)
    ## Reject some inadequate data points  
    ## (When the uniform proposal density is used, the normalizing constant in 
    ## front of the vMF density has no effects in rejection sampling.)
    mu = mu.reshape(d,1)
    sams = sam_can[unif_sam < np.exp(kappa*(np.dot(sam_can, mu)-1))[:,0],:]
    cnt = sams.shape[0]
    data_ps[:cnt,:] = sams
    while cnt < n:
        can_p = np.random.multivariate_normal(mean=np.zeros((d,)), cov=np.identity(d), size=1)
        can_p = can_p/np.sqrt(np.sum(can_p**2))
        unif_p = np.random.uniform(0, 1, 1)
        if np.exp(kappa*(np.dot(can_p, mu)-1)) > unif_p:
            data_ps[cnt,:] = can_p
            cnt += 1
    return data_ps


def vMFMixtureSamp(n, mu=np.array([[0,0,1]]), kappa=[1.0], prob=[1.0]):
    '''
    Randomly sampling data points from a mixture of q-dimensional von-Mises 
    Fisher (vMF) densities.
    
    Parameters
    ---------
        n: int
            The number of sampling random data points.
    
        mu: (m,d)-array
            The Euclidean coordinates of the m mean directions for a mixture of 
            vMF densities.
       
        kappa: list of floats
            The concentration parameters for the mixture of von-Mises Fisher 
            densities.
       
        prob: list of floats
            The mixture probabilities.
            
    Return
    --------
        data_ps: (n,d)-array
            The Euclidean coordinates of the randomly sampled points from the 
            vMF mixture.
    '''
    m = len(prob)   ## The number of mixtures
    d = mu.shape[1]  ## Euclidean dimension of the data
    assert (len(kappa) == len(prob)), "The parameters 'kappa' and 'prob' should be of the same length."
    inds = np.random.choice(list(range(m)), n, replace=True, p=np.array(prob)/sum(prob))
    data_ps = np.zeros((n,d))
    for i in range(m):
        data_ps[inds == i,:] = vMFSamp(sum(inds == i), mu=mu[i,:], kappa=kappa[i])
    return data_ps


def vMFGaussMixture(n, q=2, D=2, mu_vMF=np.array([[0,0,1]]), kappa=[1.0], 
                    mu_N=np.array([[1,1]]), cov=np.diag([1,1]).reshape(2,2,1), 
                    prob=[1.0]):
    '''
    Randomly sampling data points from a mixture of q-dimensional von-Mises Fisher 
    and D-dimensional Gaussian distributions (directional-linear mixture model).
    
    Parameters
    -----------
        n: int
            The number of sampling random data points.
            
        q: int
            Intrinsic data dimension of directional components.
            
        D: int
            Data dimension of linear components.
    
        mu_vMF: a (m,q+1)-array
            Euclidean coordinates of the m mean directions for the mixture of 
            von-Mises Fisher densities.
       
        kappa: list of floats
            The m concentration parameters for the mixture of von-Mises Fisher 
            densities.
            
        mu_N: (m,D)-array
            The means of the Gaussian mixture model with m components. 
       
        cov: (D,D,m)-array
            The (D,D)-covariance matrices of the Gaussian mixture model with 
            m components.
       
        prob: list of floats
            The m mixture probabilities.
            
    Return
    -------
        data_ps: (n, q+1+D)-array
            Euclidean coordinates of the randomly sampled points from the 
            vMF-Gaussian mixtures.
    '''
    m = len(prob)   ## The number of mixtures
    
    assert (len(kappa) == len(prob)), "The parameters 'kappa' and 'prob' "\
    "should be of the same length."
    assert (cov.shape[2] == len(prob)), "'cov.shape[2]' and 'len(prob)' "\
    "should be equal."
    
    inds = np.random.choice(list(range(m)), n, replace=True, 
                            p=np.array(prob)/sum(prob))
    data_ps = np.zeros((n,q+1+D))
    for i in range(m):
        data_ps[inds == i,:(q+1)] = vMFSamp(sum(inds == i), mu=mu_vMF[i,:], 
                                             kappa=kappa[i])
        data_ps[inds == i,(q+1):(q+1+D)] \
            = np.random.multivariate_normal(mu_N[i,:], cov[:,:,i], size=sum(inds == i))
    return data_ps


def SmoothBootstrap_vMF(data, B=1000, h=None):
    '''
    Resampling a dataset using the smoothed bootstrap with the von Mises kernel.
    
    Parameters
    ---------
        data: (n,d)-array
            The Euclidean coordinates of n directional data points in 
            the d-dimensional Euclidean space, where d=q+1.
            
        B: int
            The number of bootstrapping times.
            
        h: float
           The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
           directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
           is applied.)
            
    Return
    --------
        data_Boot: (n,d)-array
            The Euclidean coordinates of the smoothed bootstrap dataset.
    '''
    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Euclidean Dimension of the data

    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (D - R_bar ** 2) / (1 - R_bar ** 2)
        if D == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(D / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (D / 2) * (2 * (D - 1) * sp.iv(D/2, 2*kap_hat) + \
                                  (D+1) * kap_hat * sp.iv(D/2+1, 2*kap_hat)))) ** (1/(D + 3))
        print("The current bandwidth is " + str(h) + ".\n")
        
    data_Boot = np.zeros((B, d))
    for i in range(B):
        ind = np.random.choice(n, size=1, replace=True)
        data_Boot[i,:] = vMFSamp(1, mu=data[ind[0],:], kappa=1/(h**2))
    return data_Boot


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Yikun Zhang
Last Editing: June 19, 2022

Description: This script contains code for the parallel implementations with 
Ray (https://www.ray.io/) of the directional kernel density estimator with the 
von Mises kernel, its gradient estimator, directional mean shift (DMS), and 
directional subspace constrained mean shift (DirSCMS) algorithms.
"""

import numpy as np
from numpy import linalg as LA
import scipy.special as sp
import ray

#==========================================================================================#

@ray.remote
def DirMS_Ray(y_0, data, h=None, eps=1e-7, max_iter=1000, wt=None, diff_method='all'):
    '''
    Directional mean shift algorithm with the von-Mises Kernel
    
    Parameters:
        y_0: (N,d)-array
            The Euclidean coordinates of N directional initial points in 
            d-dimensional Euclidean space.
    
        data: (n,d)-array
            The Euclidean coordinates of n directional random sample points in 
            d-dimensional Euclidean space.
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
        
        eps: float
            The precision parameter for stopping the mean shift iteration.
            (Default: eps=1e-7)
        
        max_iter: int
            The maximum number of iterations for the mean shift iteration.
            (Default: max_iter=1000)
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has the 
            weight "1/n".)
            
        diff_method: str ('all'/'mean')
            The method of computing the differences between two consecutive sets
            of iteration points when they are compared with the precision 
            parameter to stop the algorithm. (When diff_method='all', all the 
            differences between two consecutive sets of iteration points need 
            to be smaller than 'eps' for terminating the algorithm. When 
            diff_method='mean', only the mean difference is compared with 'eps'
            and stop the algorithm. Default: diff_method='all'.)
    
    Return:
        MS_new: (m,d)-array
            The collection of converged points yielded by the DMS algorithm.
    '''

    n = data.shape[0]  ## Number of data points
    d = data.shape[1]  ## Euclidean dimension of the data

    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (d - R_bar ** 2) / (1 - R_bar ** 2)
        if d == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - \
                  2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(d / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (d / 2) * (2 * (d - 1) * sp.iv(d/2, 2*kap_hat) + \
                                  (d+1) * kap_hat * sp.iv(d/2+1, 2*kap_hat)))) ** (1/(d + 3))
        print("The current bandwidth is " + str(h) + ".\n")

    MS_new = np.copy(y_0)
    if wt is None:
        wt = np.ones((n,))
    for t in range(1, max_iter):
        # y_can = np.dot(np.exp(np.dot(MS_path[:,:,t-1], data.T/(h**2))), data)
        MS_old = np.copy(MS_new)
        y_can = np.dot(np.exp((np.dot(MS_old, data.T)-1)/(h**2)), data * wt.reshape(n,1))
        y_dist = np.sqrt(np.sum(y_can ** 2, axis=1))
        y_can = y_can[y_dist != 0,:]
        y_dist = y_dist[y_dist != 0]
        MS_new = y_can / y_dist.reshape(len(y_dist), 1)
        if diff_method == 'mean' and \
        np.mean(1- np.diagonal(np.dot(MS_new, MS_old.T))) <=eps:
            break
        else:
            if all(1 - np.diagonal(np.dot(MS_new, MS_old.T)) <= eps):
                break       
    '''
    if t < max_iter:
        print('The directional mean shift algorithm converges in ' + str(t) + ' steps!')
    else:
        print('The directional mean shift algorithm reaches the maximum number '\
              'of iterations,' + str(max_iter) + ' and has not yet converged.')
    '''
    return MS_new[~np.isnan(MS_new[:,0])]


@ray.remote
def DirSCMSLog_Ray(mesh_0, data, d=1, h=None, eps=1e-7, max_iter=1000, wt=None,
                   stop_cri='proj_grad'):
    '''
    Directional Subspace Constrained Mean Shift algorithm with the log density 
    and von Mises kernel
    
    Parameters:
        mesh_0: a (m,D)-array
            The Euclidean coordinates of m directional initial points in the 
            D-dimensional Euclidean space.
    
        data: a (n,D)-array
            The Euclidean coordinates of n directional data sample points in the 
            D-dimensional Euclidean space.
       
        d: int
            The order of the density ridge. (Default: d=1.)
       
        h: float
            The bandwidth parameter. (Default: h=None. Then a rule of thumb for 
            directional KDEs with the von Mises kernel in Garcia-Portugues (2013)
            is applied.)
       
        eps: float
            The precision parameter. (Default: eps=1e-7.)
       
        max_iter: int
            The maximum number of iterations for the directional SCMS algorithm 
            on each initial point. (Default: max_iter=1000.)
            
        wt: (n,)-array
            The weights of kernel density contributions for n directional random 
            sample points. (Default: wt=None, that is, each data point has an equal
            weight "1/n".)
            
        stop_cri: string ('proj_grad'/'pts_diff')
            The indicator of which stopping criteria that will be used to 
            terminate the SCMS algorithm. (When stop_cri='pts_diff', the errors 
            between two consecutive iteration points need to be smaller than 
            'eps' for terminating the algorithm. When stop_cri='proj_grad' or 
            others, the projected/principal (Riemannian) gradient of the current 
            point need to be smaller than 'eps' for terminating the algorithm.)
            (Default: stop_cri='proj_grad'.)
    
    Return:
        SCMS_new: (m,d)-array
            The collection of converged points yielded by the DirSCMS algorithm.
    '''
    
    n = data.shape[0]  ## Number of data points
    D = data.shape[1]  ## Euclidean dimension of data points
    ## Rule of thumb for directional KDE
    if h is None:
        R_bar = np.sqrt(sum(np.mean(data, axis=0) ** 2))
        ## An approximation to kappa (Banerjee 2005 & Sra, 2011)
        kap_hat = R_bar * (D - R_bar ** 2) / (1 - R_bar ** 2)
        if D == 3:
            h = (8*np.sinh(kap_hat)**2/(n*kap_hat * \
                 ((1+4*kap_hat**2)*np.sinh(2*kap_hat) - 2*kap_hat*np.cosh(2*kap_hat))))**(1/6)
        else:
            h = ((4 * np.sqrt(np.pi) * sp.iv(D / 2 - 1, kap_hat)**2) / \
                 (n * kap_hat ** (D / 2) * (2 * (D - 1) * sp.iv(D/2, 2*kap_hat) + \
                                  (D+1) * kap_hat * sp.iv(D/2+1, 2*kap_hat)))) ** (1/(D + 3))
        print("The current bandwidth is " + str(h) + ".\n")
    
    SCMS_new = np.copy(mesh_0)
    
    ## Create a vector indicating the convergent status of every mesh point
    conv_sign = np.zeros((mesh_0.shape[0], ))  
    if wt is None:
        wt = np.ones((n,1))
    else:
        wt = n*wt.reshape(n,1)
    for t in range(1, max_iter):
        if all(conv_sign == 1):
            # print('The directional SCMS algorithm converges in ' + str(t-1) + 'steps!')
            break
        SCMS_old = np.copy(SCMS_new)
        for i in range(mesh_0.shape[0]):
            if conv_sign[i] == 0:
                x_pts = SCMS_old[i,:]
                x_pts = x_pts.reshape(len(x_pts), 1)
                ## Compute the directional KDE up to a constant
                den_prop = np.sum(wt*np.exp((np.dot(data, x_pts)-1)/(h**2)))
                if den_prop == 0:
                    # Set those points with zero density values to NaN
                    nan_arr = np.zeros((x_pts.shape[0], ))
                    nan_arr[:] = np.nan
                    conv_sign[i] = 1
                    x_new = nan_arr
                else:
                    ## Compute the total gradient of the log density
                    vtot_Log_grad = np.sum(wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)), axis=0) \
                                    / ((h**2)*den_prop)
                    ## Compute the Hessian of the log density 
                    Log_Hess = np.dot(data.T, wt*data*np.exp((np.dot(data, x_pts)-1)/(h**2)))/((h**4)*den_prop) \
                               - np.dot(vtot_Log_grad.reshape(D,1), vtot_Log_grad.reshape(1,D)) \
                               - np.diag(np.sum(np.dot(wt*data, x_pts) * np.exp((np.dot(data, x_pts)-1)/(h**2))) \
                                         * np.ones(D,))/((h**2)*den_prop)
                    proj_mat = np.diag(np.ones(D,)) - np.dot(x_pts, x_pts.T)
                    Log_Hess = np.dot(np.dot(proj_mat, Log_Hess), proj_mat)
                    w, v = LA.eig(Log_Hess)
                    ## Obtain the eigenpairs inside the tangent space
                    tang_eig_v = v[:, (abs(np.dot(x_pts.T, v)) < 1e-8)[0,:]]
                    tang_eig_w = w[(abs(np.dot(x_pts.T, v)) < 1e-8)[0,:]]
                    V_d = tang_eig_v[:, np.argsort(tang_eig_w)[:(x_pts.shape[0]-1-d)]]
                    ## Iterative vector for the directional mean shift algorithm
                    ms_v = vtot_Log_grad / LA.norm(vtot_Log_grad)
                    ## Subspace constrained gradient and mean shift vector
                    SCMS_grad = np.dot(V_d, np.dot(V_d.T, vtot_Log_grad))
                    SCMS_v = np.dot(V_d, np.dot(V_d.T, ms_v))
                    ## SCMS update
                    x_new = SCMS_v + x_pts.reshape(x_pts.shape[0], )
                    x_new = x_new / LA.norm(x_new)
                    ## Stopping criteria
                    if stop_cri == 'pts_diff':
                        if LA.norm(x_new - x_pts) < eps:
                            conv_sign[i] = 1
                    else: 
                        if LA.norm(SCMS_grad) < eps:
                            conv_sign[i] = 1
                    if sum(x_new.imag) > 0:
                        # Remove those points with nonzero imaginary parts
                        nan_arr = np.zeros(x_pts.shape[0], )
                        nan_arr[:] = np.nan
                        conv_sign[i] = 1
                        x_new = nan_arr
                SCMS_new[i,:] = x_new.real
            else:
                SCMS_new[i,:] = SCMS_old[i,:]
        # print(t)
    
    '''
    if t >= max_iter-1:
        print('The directional SCMS algorithm reaches the maximum number of '\
              'iterations,'+str(max_iter)+', and has not yet converged.')
    '''
    SCMS_new = SCMS_new[conv_sign == 1]
    # Drop NaN entries before returning the ridge candidates
    return SCMS_new[~np.isnan(SCMS_new[:,0])]



#!/usr/bin/env python

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sconce-scms",
    version="0.1.2",
    author="Yikun Zhang",
    author_email="yikunzhang@foxmail.com",
    description="Spherical and Conic Cosmic Web Finders with Extended SCMS Algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zhangyk8/sconce-scms",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(include=["sconce"]),
    install_requires=["numpy", "scipy", "ray[default]", "scikit-learn"],
    python_requires=">=3.6",
)
